#!/usr/bin/env python

import sys
import getopt
import math
import time
import struct
from pymodbus.client.sync import ModbusSerialClient

endpoint = '/dev/ttyS1'
slave_addr = 0x31  # 0x31 = 49

verbose = False


def bcrc(a, n1, n2):
    assert type(a) == int
    assert 0 <= n1 <= n2 and n2 >= 0

    a = a >> n1
    crc = a & 0x01
    for i in range(n1, n2 + 1):
        a = a >> 1
        crc = crc ^ (a & 0x01)
    return crc


def wgaddc(a, n):
    assert type(a) == int
    assert n >= 24
    c0 = bcrc(a, 0, n / 2 - 1)
    c1 = ((~bcrc(a, n / 2, n - 1)) & 0x01) << (n + 1)
    return ((a << 1) | c0) | c1


def bmask(n):
    if n < 1:
        return 0
    r = 0x01
    n -= 1
    while n > 0:
        r = r << 1 | 0x01
        n -= 1
    return r


def bkey(s, l, alphabet, a_size):
    """
    :type s: str
    :param s: plate_text
    :type l: int
    :type l: max-bits
    :type a_size: int
    :param alphabet: dict
    :return:
    """
    bstr = 0
    power = 1
    for c in s.decode('utf-8').upper():
        d = ord(c)
        bts = alphabet.get(d, [0xFF, 0xFF])
        if bts and bts[0] < 0xFF:
            for i in range(2):
                if bts[i] < 0xFF:
                    bstr += bts[i] * power
                    power *= a_size
                    if verbose: print("%s=%s: %d: %d ---> %d" % (c, d, i, bts[i], bstr))
        else:
            dd = d
            while dd > a_size * a_size:
                b = dd % a_size
                bstr += b * power
                power *= a_size
                dd /= a_size

    mm = math.ceil(math.log(bstr) / math.log(2))
    while True:
        m = math.ceil(math.log(bstr) / math.log(2))
        if m > l:
            k = m % l if m < 2 * l else l
            bstr = int((bstr ^ ((bstr & bmask(int(k))) << int(m - k))) >> int(k))
        else:
            break
    return (bstr, mm)


def unpack_alphabet(filepath):
    """
    :type filepath: str
    :return alphabet table and alphabet size
    :rtype: (dict(list), int)
    """
    with open(filepath, 'rb') as bf:
        file_data = bf.read()

    reply = dict()
    offset = 0
    unique_1_sym = set()

    while offset < len(file_data):
        ucp = struct.unpack_from('>I', file_data, offset)
        offset += 4
        items = struct.unpack_from('BB', file_data, offset)
        offset += 2

        if items[1] == 0xFF:
            unique_1_sym.add(items[0])

        reply[ucp[0]] = items

    return reply, len(unique_1_sym)


def usage(msg=None):
    if msg:
        print("ERROR: " + str(msg))
    print("""
        platekey.py 
            -h              help
            -k              print key string for controller
            -w [0,1]        write to wiegand port default=none
            --wiegand=
            -b [24,32,48]   key length default=24 (wiegand: 26,34,50)      
            --bits=
            -p [plate]      plate to hash
            --plate=
            -f [file]       path to bin file with alphabet
            --binfile=
    """)
    sys.exit(1)


def trykey(p, b, w, alphabet, a_size):
    _m = 8
    (bk, bm) = bkey(p, b, alphabet, a_size)
    k = wgaddc(bk, b)
    _wds = int(math.ceil((b + 2) / 16.0))
    k <<= (_wds * 16 - b - 2)
    r = []
    for _w in range(_wds):
        r.insert(0, k & 0xffff)
        k >>= 16

    with mi as modbus:
        modbus.write_register(w, b + 2, unit=slave_addr)
        time.sleep(0.5)
        modbus.write_registers(0 if w == 0 else 0x20, r, unit=slave_addr)


def keyform(p, b, alphabet, a_size):
    (bk, bm) = bkey(p, b, alphabet, a_size)
    return "%014X" % ((bk << 8 | 0x01) & bmask(64)), bm


if __name__ == "__main__":
    options, args = getopt.getopt(sys.argv[1:], 'hHkvw:b:p:f:',
                                  ['help', 'key', 'verbose', 'wiegand=', 'bits=', 'plate=', 'binfile='])

    keyonly = False
    plate = None
    wiegand = None
    bits = 24
    bin_filepath = './def-charset.bin'

    try:
        for opt, value in options:
            if opt in ('-h', '-H', '--help'):
                usage()
            if opt in ('-v'):
                verbose = True
            if opt in ('-k', '--key'):
                keyonly = True
            if opt in ('-w', '--wiegand'):
                wiegand = int(value)
            if opt in ('-b', '--bits'):
                bits = int(value)
            if opt in ('-p', '--plate'):
                plate = value
            if opt in ('-f', '--binfile'):
                bin_filepath = value

    except getopt.error, msg:
        usage(msg)

    if not bin_filepath:
        usage("File is mandatory")

    if not plate:
        usage("Plate is mandatory")

    alp, alp_size = unpack_alphabet(bin_filepath)

    if keyonly:
        (kf, bm) = keyform(plate, bits, alp, alp_size)
        print(kf)
        if verbose: print(int(bm))
        sys.exit(0)
    if wiegand is None:
        usage("Wiegand port is mandatory")

    mi = ModbusSerialClient(method="rtu", port=endpoint, baudrate=19200, bytesize=8, stopbits=1, parity='N')
    mi.timeout = 1

    trykey(plate, bits, wiegand, alp, alp_size)
    sys.exit(0)
